# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle

import pyomo.common.unittest as unittest
from pyomo.core.expr.numvalue import (
    NumericValue,
    is_fixed,
    is_constant,
    is_potentially_variable,
)
from pyomo.core.tests.unit.kernel.test_dict_container import (
    _TestActiveDictContainerBase,
)
from pyomo.core.tests.unit.kernel.test_tuple_container import (
    _TestActiveTupleContainerBase,
)
from pyomo.core.tests.unit.kernel.test_list_container import (
    _TestActiveListContainerBase,
)
from pyomo.core.kernel.base import ICategorizedObject
from pyomo.core.kernel.parameter import parameter
from pyomo.core.kernel.variable import (
    IVariable,
    variable,
    variable_dict,
    variable_tuple,
    variable_list,
    _extract_domain_type_and_bounds,
)
from pyomo.core.kernel.block import block
from pyomo.core.kernel.set_types import RealSet, IntegerSet, BooleanSet
from pyomo.core.base.set import (
    Binary,
    NonNegativeReals,
    NegativeReals,
    Reals,
    NonNegativeIntegers,
    NegativeIntegers,
    RealInterval,
    IntegerInterval,
)


class Test_variable(unittest.TestCase):
    def test_pprint(self):
        import pyomo.kernel

        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        v = variable()
        pyomo.kernel.pprint(v)
        b = block()
        b.v = v
        pyomo.kernel.pprint(v)
        pyomo.kernel.pprint(b)
        m = block()
        m.b = b
        pyomo.kernel.pprint(v)
        pyomo.kernel.pprint(b)
        pyomo.kernel.pprint(m)

    def test_bad_bounds(self):
        v = variable()
        self.assertIs(v.lb, None)
        self.assertIs(v.ub, None)
        v.lb = 1.0
        v.ub = 1.0
        self.assertEqual(v.lb, 1.0)
        self.assertEqual(v.ub, 1.0)
        with self.assertRaises(ValueError):
            v.lb = "string"
        self.assertEqual(v.lb, 1.0)
        self.assertEqual(v.ub, 1.0)
        with self.assertRaises(ValueError):
            v.ub = "string"
        self.assertEqual(v.lb, 1.0)
        self.assertEqual(v.ub, 1.0)

    def test_extract_domain_type_and_bounds(self):
        # test an edge case
        domain_type, lb, ub = _extract_domain_type_and_bounds(None, None, None, None)
        self.assertIs(domain_type, RealSet)
        self.assertIs(lb, None)
        self.assertIs(ub, None)

    def test_polynomial_degree(self):
        v = variable()
        self.assertEqual(v.polynomial_degree(), 1)
        v.fix(0)
        self.assertEqual(v.polynomial_degree(), 0)
        v.free()
        self.assertEqual(v.polynomial_degree(), 1)

    def test_ctype(self):
        v = variable()
        self.assertIs(v.ctype, IVariable)
        self.assertIs(type(v), variable)
        self.assertIs(type(v)._ctype, IVariable)

    def test_pickle(self):
        v = variable(lb=1, ub=2, domain_type=IntegerSet, fixed=True)
        self.assertEqual(v.lb, 1)
        self.assertEqual(type(v.lb), int)
        self.assertEqual(v.ub, 2)
        self.assertEqual(type(v.ub), int)
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.fixed, True)
        self.assertEqual(v.parent, None)
        vup = pickle.loads(pickle.dumps(v))
        self.assertEqual(vup.lb, 1)
        self.assertEqual(type(vup.lb), int)
        self.assertEqual(vup.ub, 2)
        self.assertEqual(type(vup.ub), int)
        self.assertEqual(vup.domain_type, IntegerSet)
        self.assertEqual(vup.fixed, True)
        self.assertEqual(vup.parent, None)
        b = block()
        b.v = v
        self.assertIs(v.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        vup = bup.v
        self.assertEqual(vup.lb, 1)
        self.assertEqual(vup.ub, 2)
        self.assertEqual(vup.domain_type, IntegerSet)
        self.assertEqual(vup.fixed, True)
        self.assertIs(vup.parent, bup)

    def test_init(self):
        v = variable()
        self.assertTrue(v.parent is None)
        self.assertEqual(v.ctype, IVariable)
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, None)
        self.assertEqual(v.stale, True)
        b = block()
        b.v = v
        self.assertTrue(v.parent is b)
        del b.v
        self.assertTrue(v.parent is None)

        v = variable(domain_type=IntegerSet, value=1, lb=0, ub=2, fixed=True)
        self.assertTrue(v.parent is None)
        self.assertEqual(v.ctype, IVariable)
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 2)
        self.assertEqual(v.fixed, True)
        self.assertEqual(v.value, 1)
        self.assertEqual(v.stale, True)

    def test_type(self):
        v = variable()
        self.assertTrue(isinstance(v, ICategorizedObject))
        self.assertTrue(isinstance(v, IVariable))
        self.assertTrue(isinstance(v, NumericValue))

    def test_is_constant(self):
        v = variable()
        self.assertEqual(v.is_constant(), False)
        self.assertEqual(is_constant(v), False)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, None)
        v.value = 1.0
        self.assertEqual(v.is_constant(), False)
        self.assertEqual(is_constant(v), False)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, 1.0)
        v.fix()
        self.assertEqual(v.is_constant(), False)
        self.assertEqual(is_constant(v), False)
        self.assertEqual(v.fixed, True)
        self.assertEqual(v.value, 1.0)
        v.value = None
        self.assertEqual(v.is_constant(), False)
        self.assertEqual(is_constant(v), False)
        self.assertEqual(v.fixed, True)
        self.assertEqual(v.value, None)
        v.free()
        self.assertEqual(v.is_constant(), False)
        self.assertEqual(is_constant(v), False)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, None)

    def test_is_fixed(self):
        v = variable()
        self.assertEqual(v.is_fixed(), False)
        self.assertEqual(is_fixed(v), False)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, None)
        v.value = 1.0
        self.assertEqual(v.is_fixed(), False)
        self.assertEqual(is_fixed(v), False)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, 1.0)
        v.fix()
        self.assertEqual(v.is_fixed(), True)
        self.assertEqual(is_fixed(v), True)
        self.assertEqual(v.fixed, True)
        self.assertEqual(v.value, 1.0)
        v.value = None
        self.assertEqual(v.is_fixed(), True)
        self.assertEqual(is_fixed(v), True)
        self.assertEqual(v.fixed, True)
        self.assertEqual(v.value, None)
        v.free()
        self.assertEqual(v.is_fixed(), False)
        self.assertEqual(is_fixed(v), False)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, None)

    def test_potentially_variable(self):
        v = variable()
        self.assertEqual(v.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(v), True)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, None)
        v.value = 1.0
        self.assertEqual(v.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(v), True)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, 1.0)
        v.fix()
        self.assertEqual(v.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(v), True)
        self.assertEqual(v.fixed, True)
        self.assertEqual(v.value, 1.0)
        v.value = None
        self.assertEqual(v.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(v), True)
        self.assertEqual(v.fixed, True)
        self.assertEqual(v.value, None)
        v.free()
        self.assertEqual(v.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(v), True)
        self.assertEqual(v.fixed, False)
        self.assertEqual(v.value, None)

    def test_domain(self):
        v = variable(domain=Reals)
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.bounds, (None, None))

        v = variable(domain=Reals, lb=0, ub=1)
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))

        lb_param = parameter()
        ub_param = parameter()
        lb = lb_param
        ub = ub_param
        v = variable(domain=Reals, lb=lb, ub=ub)
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertIs(v.lower, lb)
        self.assertIs(v.upper, ub)
        with self.assertRaisesRegex(
            ValueError, 'No value for uninitialized parameter object None'
        ):
            v.bounds

        lb = lb**2
        ub = ub**2
        v.lb = lb
        v.ub = ub
        lb_param.value = 2.0
        ub_param.value = 3.0
        self.assertIs(v.lower, lb)
        self.assertEqual(v.lb, 4.0)
        self.assertIs(v.upper, ub)
        self.assertEqual(v.ub, 9.0)
        self.assertEqual(v.bounds[0], 4.0)
        self.assertEqual(v.bounds[1], 9.0)

        # if the domain has a finite bound
        # then you can not use the same bound
        # keyword
        variable(domain=NonNegativeReals, ub=0)
        with self.assertRaises(ValueError):
            variable(domain=NonNegativeReals, lb=0)

        variable(domain=NonNegativeReals, ub=ub)
        with self.assertRaises(ValueError):
            variable(domain=NonNegativeReals, lb=lb)

        variable(domain=NonNegativeIntegers, ub=0)
        with self.assertRaises(ValueError):
            variable(domain=NonNegativeIntegers, lb=0)

        variable(domain=NonNegativeIntegers, ub=ub)
        with self.assertRaises(ValueError):
            variable(domain=NonNegativeIntegers, lb=lb)

        variable(domain=NegativeReals, lb=0)
        with self.assertRaises(ValueError):
            variable(domain=NegativeReals, ub=0)

        variable(domain=NegativeReals, lb=lb)
        with self.assertRaises(ValueError):
            variable(domain=NegativeReals, ub=ub)

        variable(domain=NegativeIntegers, lb=0)
        with self.assertRaises(ValueError):
            variable(domain=NegativeIntegers, ub=0)

        variable(domain=NegativeIntegers, lb=lb)
        with self.assertRaises(ValueError):
            variable(domain=NegativeIntegers, ub=ub)

        unit_interval = RealInterval(bounds=(0, 1))
        self.assertEqual(unit_interval.bounds(), (0, 1))
        v = variable(domain=unit_interval)
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))
        with self.assertRaises(ValueError):
            variable(domain=unit_interval, lb=0)
        with self.assertRaises(ValueError):
            variable(domain=unit_interval, ub=0)

        v = variable()
        v.domain = unit_interval
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))

        binary = IntegerInterval(bounds=(0, 1))
        self.assertEqual(binary.bounds(), (0, 1))
        v = variable(domain=binary)
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))
        with self.assertRaises(ValueError):
            variable(domain=binary, lb=0)
        with self.assertRaises(ValueError):
            variable(domain=binary, ub=0)

        v = variable()
        v.domain = binary
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))

        variable(domain_type=RealSet)
        variable(domain=Reals)
        with self.assertRaises(ValueError):
            variable(domain_type=RealSet, domain=Reals)
        with self.assertRaises(ValueError):
            variable(domain_type=BooleanSet)

    def test_domain_type(self):
        v = variable()
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.bounds, (None, None))

        v.domain_type = IntegerSet
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.bounds, (None, None))

        v = variable(domain_type=IntegerSet, lb=0)
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.bounds, (0, None))

        v = variable(domain_type=IntegerSet, ub=0)
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, 0)
        self.assertEqual(v.bounds, (None, 0))

        v.domain_type = RealSet
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, 0)
        self.assertEqual(v.bounds, (None, 0))

        with self.assertRaises(ValueError):
            v.domain_type = BooleanSet

    def test_binary_type(self):
        v = variable()
        v.domain_type = IntegerSet
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.bounds, (None, None))
        self.assertEqual(v.has_lb(), False)
        self.assertEqual(v.has_ub(), False)

        v.lb = 0
        v.ub = 1
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(type(v.lb), int)
        self.assertEqual(v.ub, 1)
        self.assertEqual(type(v.ub), int)
        self.assertEqual(v.bounds, (0, 1))

        v.lb = 0
        v.ub = 0
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 0)
        self.assertEqual(v.bounds, (0, 0))

        v.lb = 1
        v.ub = 1
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 1)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (1, 1))

        v = variable(domain=Binary)
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))

        v.ub = 2
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 2)
        self.assertEqual(v.bounds, (0, 2))

        v.lb = -1
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, -1)
        self.assertEqual(v.ub, 2)
        self.assertEqual(v.bounds, (-1, 2))

        v.domain = Binary
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))

        v.domain_type = RealSet
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))

        v.domain_type = IntegerSet
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))

        v.domain = Reals
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.bounds, (None, None))
        self.assertEqual(v.has_lb(), False)
        self.assertEqual(v.has_ub(), False)

        v.bounds = (float('-inf'), float('inf'))
        self.assertEqual(v.domain_type, RealSet)
        self.assertEqual(v.is_continuous(), True)
        self.assertEqual(v.is_discrete(), False)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), False)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.bounds, (None, None))
        self.assertEqual(v.has_lb(), False)
        self.assertEqual(v.has_ub(), False)

        pL = parameter()
        v.lb = pL
        pU = parameter()
        v.ub = pU

        v.domain_type = IntegerSet
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        with self.assertRaises(ValueError):
            v.is_binary()
        self.assertEqual(v.is_integer(), True)
        self.assertIs(v.lower, pL)
        with self.assertRaises(ValueError):
            v.has_lb()
        self.assertIs(v.upper, pU)
        with self.assertRaises(ValueError):
            v.has_ub()

        pL.value = 0
        pU.value = 1
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.has_lb(), True)
        self.assertEqual(v.has_ub(), True)

        pL.value = 0
        pU.value = 0
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 0)

        pL.value = 1
        pU.value = 1
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), True)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 1)
        self.assertEqual(v.ub, 1)

        v.domain = Binary
        pU.value = 2
        v.ub = pU
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 2)

        pL.value = -1
        v.lb = pL
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, -1)
        self.assertEqual(v.ub, 2)

        pL.value = float('-inf')
        pU.value = float('inf')
        self.assertEqual(v.domain_type, IntegerSet)
        self.assertEqual(v.is_continuous(), False)
        self.assertEqual(v.is_discrete(), True)
        self.assertEqual(v.is_binary(), False)
        self.assertEqual(v.is_integer(), True)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.ub, None)
        self.assertEqual(v.has_lb(), False)
        self.assertEqual(v.has_ub(), False)

    def test_bounds_setter(self):
        v = variable()
        v.lb = 0
        v.ub = 1
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.ub, 1)
        self.assertEqual(v.bounds, (0, 1))

        v.bounds = (2, 3)
        self.assertEqual(v.lb, 2)
        self.assertEqual(v.ub, 3)
        self.assertEqual(v.bounds, (2, 3))

        v.lb = -1
        v.ub = 0
        self.assertEqual(v.lb, -1)
        self.assertEqual(v.ub, 0)
        self.assertEqual(v.bounds, (-1, 0))

    def test_has_lb_ub(self):
        v = variable()
        self.assertEqual(v.has_lb(), False)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.has_ub(), False)
        self.assertEqual(v.ub, None)

        v.lb = float('-inf')
        self.assertEqual(v.has_lb(), False)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.has_ub(), False)
        self.assertEqual(v.ub, None)

        v.ub = float('inf')
        self.assertEqual(v.has_lb(), False)
        self.assertEqual(v.lb, None)
        self.assertEqual(v.has_ub(), False)
        self.assertEqual(v.ub, None)

        v.lb = 0
        self.assertEqual(v.has_lb(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.has_ub(), False)
        self.assertEqual(v.ub, None)

        v.ub = 0
        self.assertEqual(v.has_lb(), True)
        self.assertEqual(v.lb, 0)
        self.assertEqual(v.has_ub(), True)
        self.assertEqual(v.ub, 0)

        #
        # edge cases
        #

        v.lb = float('inf')
        self.assertEqual(v.has_lb(), True)
        self.assertEqual(v.lb, float('inf'))
        self.assertEqual(v.has_ub(), True)
        self.assertEqual(v.ub, 0)

        v.ub = float('-inf')
        self.assertEqual(v.has_lb(), True)
        self.assertEqual(v.lb, float('inf'))
        self.assertEqual(v.has_ub(), True)
        self.assertEqual(v.ub, float('-inf'))

    def test_fix_free(self):
        v = variable()
        self.assertEqual(v.value, None)
        self.assertEqual(v.fixed, False)

        v.fix(1)
        self.assertEqual(v.value, 1)
        self.assertEqual(v.fixed, True)

        v.free()
        self.assertEqual(v.value, 1)
        self.assertEqual(v.fixed, False)

        v.value = 0
        self.assertEqual(v.value, 0)
        self.assertEqual(v.fixed, False)

        v.fix()
        self.assertEqual(v.value, 0)
        self.assertEqual(v.fixed, True)

        with self.assertRaises(TypeError):
            v.fix(1, 2)
        self.assertEqual(v.value, 0)
        self.assertEqual(v.fixed, True)

        v.free()
        with self.assertRaises(TypeError):
            v.fix(1, 2)
        self.assertEqual(v.value, 0)
        self.assertEqual(v.fixed, False)

    def test_active(self):
        v = variable()
        b = block()
        self.assertEqual(b.active, True)
        b.deactivate()
        self.assertEqual(b.active, False)
        b.v = v
        self.assertEqual(b.active, False)
        b.activate()
        self.assertEqual(b.active, True)

    def test_call(self):
        v = variable()
        self.assertEqual(v.value, None)
        with self.assertRaises(ValueError):
            v()
        with self.assertRaises(ValueError):
            v(exception=True)
        self.assertEqual(v(exception=False), None)

        v.value = 2
        self.assertEqual(v.value, 2)
        self.assertEqual(v(), 2)
        self.assertEqual(v(exception=True), 2)
        self.assertEqual(v(exception=False), 2)

    def test_slack_methods(self):
        x = variable(value=2)
        L = 1
        U = 5

        # equality
        x.bounds = L, L
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, -3)
        self.assertEqual(x.lslack, 3)
        self.assertEqual(x.uslack, -3)
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, -5)
        self.assertEqual(x.lslack, 5)
        self.assertEqual(x.uslack, -5)
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, -1)
        self.assertEqual(x.uslack, 1)
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # equality
        x.bounds = U, U
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, -1)
        self.assertEqual(x.uslack, 1)
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, 1)
        self.assertEqual(x.uslack, -1)
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, -5)
        self.assertEqual(x.lslack, -5)
        self.assertEqual(x.uslack, 5)
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # lower finite
        x.bounds = L, None
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, 3)
        self.assertEqual(x.lslack, 3)
        self.assertEqual(x.uslack, float('inf'))
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, 5)
        self.assertEqual(x.lslack, 5)
        self.assertEqual(x.uslack, float('inf'))
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, -1)
        self.assertEqual(x.uslack, float('inf'))
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # lower unbounded
        x.bounds = float('-inf'), None
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # upper finite
        x.bounds = None, U
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, 1)
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, 1)
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, -1)
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, 5)
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, 5)
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # upper unbounded
        x.bounds = None, float('inf')
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # range finite
        x.bounds = L, U
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, 1)
        self.assertEqual(x.lslack, 3)
        self.assertEqual(x.uslack, 1)
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, 5)
        self.assertEqual(x.uslack, -1)
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, -1)
        self.assertEqual(x.uslack, 5)
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # range unbounded (None)
        x.bounds = None, None
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # range unbounded
        x.bounds = float('-inf'), float('inf')
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # range finite (parameter)
        x.bounds = parameter(L), parameter(U)
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, 1)
        self.assertEqual(x.lslack, 3)
        self.assertEqual(x.uslack, 1)
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, 5)
        self.assertEqual(x.uslack, -1)
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, -1)
        self.assertEqual(x.lslack, -1)
        self.assertEqual(x.uslack, 5)
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)

        # range unbounded (parameter)
        x.bounds = parameter(float('-inf')), parameter(float('inf'))
        x.value = 4
        self.assertEqual(x.value, 4)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 6
        self.assertEqual(x.value, 6)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = 0
        self.assertEqual(x.value, 0)
        self.assertEqual(x.slack, float('inf'))
        self.assertEqual(x.lslack, float('inf'))
        self.assertEqual(x.uslack, float('inf'))
        x.value = None
        self.assertEqual(x.value, None)
        self.assertEqual(x.slack, None)
        self.assertEqual(x.lslack, None)
        self.assertEqual(x.uslack, None)


class _variable_subclass(variable):
    pass


class Test_variable_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = variable_dict
    _ctype_factory = lambda self: variable()


class Test_variable_tuple(_TestActiveTupleContainerBase, unittest.TestCase):
    _container_type = variable_tuple
    _ctype_factory = lambda self: variable()


class Test_variable_list(_TestActiveListContainerBase, unittest.TestCase):
    _container_type = variable_list
    _ctype_factory = lambda self: variable()


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import pandas as pd
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import (
    RooneyBieglerExperiment,
)


def main():

    # Data
    data = pd.DataFrame(
        data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
        columns=['hour', 'y'],
    )

    # Sum of squared error function
    def SSE(model):
        expr = (model.experiment_outputs[model.y] - model.y) ** 2
        return expr

    # Create an experiment list
    exp_list = []
    for i in range(data.shape[0]):
        exp_list.append(RooneyBieglerExperiment(data.loc[i, :]))

    # View one model
    # exp0_model = exp_list[0].get_labeled_model()
    # exp0_model.pprint()

    # Create an instance of the parmest estimator
    pest = parmest.Estimator(exp_list, obj_function=SSE)

    # Parameter estimation
    obj, theta = pest.theta_est()

    # Parameter estimation with bootstrap resampling
    bootstrap_theta = pest.theta_est_bootstrap(50, seed=4581)

    # Plot results
    parmest.graphics.pairwise_plot(bootstrap_theta, title='Bootstrap theta')
    parmest.graphics.pairwise_plot(
        bootstrap_theta,
        theta,
        0.8,
        ['MVN', 'KDE', 'Rect'],
        title='Bootstrap theta with confidence regions',
    )


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.config import ConfigDict, ConfigValue
from pyomo.common.modeling import unique_component_name
from pyomo.core import (
    Any,
    Block,
    Constraint,
    NonNegativeIntegers,
    SortComponents,
    value,
    Var,
)
from pyomo.core.base import Transformation, TransformationFactory
from pyomo.core.expr import identify_variables
from pyomo.core.util import target_list
from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.gdp.util import is_child_of, get_gdp_tree
from pyomo.repn.standard_repn import generate_standard_repn
import logging

logger = logging.getLogger(__name__)


@TransformationFactory.register(
    'gdp.bound_pretransformation',
    doc="Partially transforms a GDP to a MIP by finding all disjunctive "
    "constraints with common left-hand sides and transforming them according "
    "to the formulation in Balas 1988",
)
class BoundPretransformation(Transformation):
    """
    Implements a special case of the transformation mentioned in [1] for
    handling disjunctive constraints with common left-hand sides (i.e.,
    Constraint bodies). Automatically detects univariate disjunctive
    Constraints (bounds or equalities involving one variable), and
    transforms them according to [1]. The transformed Constraints are
    deactivated, but the remainder of the GDP is untouched. That is,
    to completely transform the GDP, a GDP-to-MIP transformation is
    needed that will transform the remaining disjunctive constraints as
    well as any LogicalConstraints and the logic of the disjunctions
    themselves.

    NOTE: Because this transformation allows tighter bound values higher in
    the GDP hierarchy to supersede looser ones that are lower, the transformed
    model will not necessarily still be valid in the case that there are
    mutable Params in disjunctive variable bounds or in the transformed
    Constraints and the values of those mutable Params are later changed.
    Similarly, if this transformation is called when Vars are fixed, it will
    only be guaranteed to be valid when those Vars remain fixed to the same
    values.

    [1] Egon Balas, "On the convex hull of the union of certain polyhedra,"
        Operations Research Letters, vol. 7, 1988, pp. 279-283
    """

    CONFIG = ConfigDict('gdp.bound_pretransformation')
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets to transform",
            doc="""
            This specifies the list of Disjunctions or Blocks to be (partially)
            transformed. If None (default), the entire model is transformed. 
            Note that if the transformation is done out of place, the list of
            targets should be attached to the model before it is cloned, and
            the list will specify the targets on the cloned instance.
            """,
        ),
    )
    transformation_name = 'bound_pretransformation'

    def __init__(self):
        super().__init__()
        self.logger = logger

    def _apply_to(self, instance, **kwds):
        if not instance.ctype in (Block, Disjunct):
            raise GDP_Error(
                "Transformation called on %s of type %s. 'instance'"
                " must be a ConcreteModel, Block, or Disjunct (in "
                "the case of nested disjunctions)." % (instance.name, instance.ctype)
            )

        self._config = self.CONFIG(kwds.pop('options', {}))
        self._config.set_value(kwds)

        targets = self._config.targets
        if targets is None:
            targets = (instance,)

        transformation_blocks = {}
        bound_dict = ComponentMap()
        self._update_bounds_from_constraints(instance, bound_dict, None, is_root=True)
        # [ESJ 05/04/23]: In the future, I should think about getting my little
        # trees from this tree, or asking for leaves rooted somewhere specific
        # or something. Because this transformation currently does the work of
        # getting the GDP tree twice...
        whole_tree = get_gdp_tree(targets, instance)
        for t in whole_tree.topological_sort():
            if t.ctype is Disjunction and whole_tree.in_degree(t) == 0:
                self._transform_disjunction(
                    t, instance, bound_dict, transformation_blocks
                )

    def _transform_disjunction(
        self, disjunction, instance, bound_dict, transformation_blocks
    ):
        # We go from root to leaves so that whenever we hit a variable, we can
        # ask if the bounds we're seeing on its parent Disjunct (or if we're at
        # the root, in the global scope) are looser or tighter than the bounds
        # on it, and we pass the tightest ones down. For sane models, the bounds
        # will tighten as we go down the tree, but that's of course not
        # guaranteed since not all models are sane...
        disjunctions_to_transform = set()
        gdp_forest = get_gdp_tree((disjunction,), instance)
        for d in gdp_forest.topological_sort():
            if d.ctype is Disjunct:
                self._update_bounds_from_constraints(d, bound_dict, gdp_forest)
        self._create_transformation_constraints(
            disjunction, bound_dict, gdp_forest, transformation_blocks
        )

    def _get_bound_dict_for_var(self, bound_dict, v):
        v_bounds = bound_dict.get(v)
        if v_bounds is None:
            v_bounds = bound_dict[v] = {
                None: (v.lb, v.ub),
                'to_deactivate': ComponentMap(),
            }
        return v_bounds

    def _update_bounds_from_constraints(
        self, disjunct, bound_dict, gdp_forest, is_root=False
    ):
        for constraint in disjunct.component_data_objects(
            Constraint,
            active=True,
            descend_into=Block,
            sort=SortComponents.deterministic,
        ):
            # Avoid walking the whole expression tree if we have more than one
            # variable by just trying to get two. If we succeed at one but not
            # two, then the constraint is a bound or equality constraint and we
            # save it. Otherwise, we just keep going to the next constraint.
            var_gen = identify_variables(constraint.body, include_fixed=False)
            try:
                next(var_gen)
            except StopIteration:
                # No variables
                continue
            try:
                next(var_gen)
            except StopIteration:
                # There was one but not two: This is what we want.
                repn = generate_standard_repn(constraint.body)
                # If this is a trivial constraint, repn could actually be empty,
                # so we check that we really do have one linear var now
                if not repn.is_linear() or len(repn.linear_vars) != 1:
                    continue
                v = repn.linear_vars[0]
                coef = repn.linear_coefs[0]
                constant = repn.constant
                lower = (
                    (value(constraint.lower) - constant) / coef
                    if constraint.lower is not None
                    else None
                )
                upper = (
                    (value(constraint.upper) - constant) / coef
                    if constraint.upper is not None
                    else None
                )
                if coef < 0:
                    # we divided by a negative coef above, so flip the constraint
                    lower, upper = (upper, lower)
                v_bounds = self._get_bound_dict_for_var(bound_dict, v)
                self._update_bounds_dict(
                    v_bounds,
                    lower,
                    upper,
                    disjunct if not is_root else None,
                    gdp_forest,
                )
                if not is_root:
                    if disjunct in v_bounds['to_deactivate']:
                        v_bounds['to_deactivate'][disjunct].add(constraint)
                    else:
                        v_bounds['to_deactivate'][disjunct] = ComponentSet([constraint])

    def _get_tightest_ancestral_bounds(self, v_bounds, disjunct, gdp_forest):
        lb = None
        ub = None
        parent = disjunct
        while lb is None or ub is None:
            if parent in v_bounds:
                l, u = v_bounds[parent]
                if lb is None and l is not None:
                    lb = l
                if ub is None and u is not None:
                    ub = u
            if parent is None:
                break
            parent = gdp_forest.parent_disjunct(parent)
        return lb, ub

    def _update_bounds_dict(self, v_bounds, lower, upper, disjunct, gdp_forest):
        lb, ub = self._get_tightest_ancestral_bounds(v_bounds, disjunct, gdp_forest)
        if lower is not None:
            if lb is None or lower > lb:
                # This GDP is more constrained here than it was in the parent
                # Disjunct (what we would expect, usually. If it's looser, we're
                # essentially just ignoring it...)
                lb = lower
        if upper is not None:
            if ub is None or upper < ub:
                # Same case as above in the UB
                ub = upper
        # In all other cases, there is nothing to do... The parent gives more
        # information, so we just propagate that down
        v_bounds[disjunct] = (lb, ub)

    def _create_transformation_constraints(
        self, root_disjunction, bound_dict, gdp_forest, transformation_blocks
    ):
        to_transform = ComponentSet([root_disjunction])

        while to_transform:
            disjunction = to_transform.pop()

            trans_block = self._add_transformation_block(
                disjunction, transformation_blocks
            )
            if self.transformation_name not in disjunction._transformation_map:
                disjunction._transformation_map[self.transformation_name] = (
                    ComponentMap()
                )
            trans_map = disjunction._transformation_map[self.transformation_name]

            for disj in disjunction.disjuncts:
                to_transform.update(gdp_forest.children(disj))

            for v, v_bounds in bound_dict.items():
                unique_id = len(trans_block.transformed_bound_constraints)
                if not any(disj in v_bounds for disj in disjunction.disjuncts):
                    # There are no bound Constraints on this Disjunction. We
                    # don't want to create a bunch of empty Blocks and things as
                    # if there were, so we continue to the next.
                    continue
                all_lbs = True
                all_ubs = True
                lb_expr = 0
                ub_expr = 0
                deactivate_lower = ComponentSet()
                deactivate_upper = ComponentSet()
                for disj in disjunction.disjuncts:
                    lb, ub = self._get_tightest_ancestral_bounds(
                        v_bounds, disj, gdp_forest
                    )
                    if lb is None:
                        all_lbs = False
                        if not all_ubs:
                            # We're not going to get all of either: we're done.
                            break
                    if ub is None:
                        all_ubs = False
                        if not all_lbs:
                            break
                    if all_lbs:
                        lb_expr += lb * disj.binary_indicator_var
                        # If these bounds came from above here in the GDP
                        # hierarchy, this disjunct might not actually have
                        # constraints to deactivate. If it does, we add them to
                        # our list to take care of if we end up being able to
                        # make a constraint
                        if disj in v_bounds['to_deactivate']:
                            deactivate_lower.update(v_bounds['to_deactivate'][disj])
                    if all_ubs:
                        ub_expr += ub * disj.binary_indicator_var
                        if disj in v_bounds['to_deactivate']:
                            deactivate_upper.update(v_bounds['to_deactivate'][disj])

                # actually make the constraint(s) now
                if all_lbs:
                    idx = (v.local_name + '_lb', unique_id)
                    trans_block.transformed_bound_constraints[idx] = lb_expr <= v
                    trans_map[v] = [trans_block.transformed_bound_constraints[idx]]
                    for c in deactivate_lower:
                        if c.lower is None:
                            # There's nothing to do
                            continue
                        if c.upper is None or (all_ubs and c in deactivate_upper):
                            c.deactivate()
                        else:
                            c.deactivate()
                            c.parent_block().add_component(
                                unique_component_name(
                                    c.parent_block(), c.local_name + '_ub'
                                ),
                                Constraint(expr=v <= c.upper),
                            )
                if all_ubs:
                    idx = (v.local_name + '_ub', unique_id + 1)
                    trans_block.transformed_bound_constraints[idx] = ub_expr >= v
                    if v in trans_map:
                        trans_map[v].append(
                            trans_block.transformed_bound_constraints[idx]
                        )
                    else:
                        trans_map[v] = [trans_block.transformed_bound_constraints[idx]]
                    for c in deactivate_upper:
                        if c.upper is None:
                            # There's nothing to do
                            continue
                        if c.lower is None or (all_lbs and c in deactivate_lower):
                            c.deactivate()
                        else:
                            c.deactivate()
                            c.parent_block().add_component(
                                unique_component_name(
                                    c.parent_block(), c.local_name + '_lb'
                                ),
                                Constraint(expr=v >= c.lower),
                            )

    def _add_transformation_block(self, disjunction, transformation_blocks):
        to_block = disjunction.parent_block()
        if to_block in transformation_blocks:
            return transformation_blocks[to_block]

        trans_block_name = unique_component_name(
            to_block, '_pyomo_gdp_%s_reformulation' % self.transformation_name
        )
        transformation_blocks[to_block] = trans_block = Block()
        to_block.add_component(trans_block_name, trans_block)

        trans_block.transformed_bound_constraints = Constraint(
            Any * NonNegativeIntegers
        )

        return trans_block

    def get_transformed_constraints(self, v, disjunction):
        if self.transformation_name not in disjunction._transformation_map:
            logger.debug(
                "No variable on Disjunction '%s' was transformed with the "
                "gdp.%s transformation" % (disjunction.name, self.transformation_name)
            )
            return []
        trans_map = disjunction._transformation_map[self.transformation_name]
        if v not in trans_map:
            logger.debug(
                "Constraint bounding variable '%s' on Disjunction '%s' was "
                "not transformed by the 'gdp.%s' transformation"
                % (v.name, disjunction.name, self.transformation_name)
            )
            return []
        return trans_map[v]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle

import pyomo.common.unittest as unittest
import pyomo.kernel as pmo
from pyomo.core.tests.unit.kernel.test_dict_container import (
    _TestActiveDictContainerBase,
)
from pyomo.core.tests.unit.kernel.test_list_container import (
    _TestActiveListContainerBase,
)
from pyomo.core.kernel.base import ICategorizedObject, ICategorizedObjectContainer
from pyomo.core.kernel.heterogeneous_container import IHeterogeneousContainer
from pyomo.core.kernel.block import IBlock, block, block_dict, block_list
from pyomo.core.kernel.variable import variable, variable_list
from pyomo.core.kernel.piecewise_library.transforms import (
    PiecewiseLinearFunction,
    TransformedPiecewiseLinearFunction,
)
import pyomo.core.kernel.piecewise_library.transforms as transforms
from pyomo.core.kernel.piecewise_library.transforms_nd import (
    PiecewiseLinearFunctionND,
    TransformedPiecewiseLinearFunctionND,
)
import pyomo.core.kernel.piecewise_library.transforms_nd as transforms_nd
import pyomo.core.kernel.piecewise_library.util as util

# for the multi-dimensional piecewise tests
_test_v = None
_test_tri = None
_test_values = None


def setUpModule():
    global _test_v
    global _test_tri
    global _test_values
    if util.numpy_available and util.scipy_available:
        _test_v = variable_list(variable(lb=i, ub=i + 1) for i in range(3))
        _test_tri = util.generate_delaunay(_test_v, num=4)
        _test_values = []
        for _xi in _test_tri.points:
            _test_values.append(sum(_xi))
        _test_values = util.numpy.array(_test_values)


class Test_util(unittest.TestCase):
    def test_is_constant(self):
        self.assertEqual(util.is_constant([]), True)
        self.assertEqual(util.is_constant([1]), True)
        self.assertEqual(util.is_constant([1, 2]), False)
        self.assertEqual(util.is_constant([1, 1]), True)
        self.assertEqual(util.is_constant([1, 2, 3]), False)
        self.assertEqual(util.is_constant([2.1, 2.1, 2.1]), True)
        self.assertEqual(util.is_constant([1, 1, 3, 4]), False)
        self.assertEqual(util.is_constant([1, 1, 3, 3]), False)
        self.assertEqual(util.is_constant([1, 1, 1, 4]), False)
        self.assertEqual(util.is_constant([1, 1, 1, 1]), True)
        self.assertEqual(util.is_constant([-1, 1, 1, 1]), False)
        self.assertEqual(util.is_constant([1, -1, 1, 1]), False)
        self.assertEqual(util.is_constant([1, 1, -1, 1]), False)
        self.assertEqual(util.is_constant([1, 1, 1, -1]), False)

    def test_is_nondecreasing(self):
        self.assertEqual(util.is_nondecreasing([]), True)
        self.assertEqual(util.is_nondecreasing([1]), True)
        self.assertEqual(util.is_nondecreasing([1, 2]), True)
        self.assertEqual(util.is_nondecreasing([1, 2, 3]), True)
        self.assertEqual(util.is_nondecreasing([1, 1, 3, 4]), True)
        self.assertEqual(util.is_nondecreasing([1, 1, 3, 3]), True)
        self.assertEqual(util.is_nondecreasing([1, 1, 1, 4]), True)
        self.assertEqual(util.is_nondecreasing([1, 1, 1, 1]), True)
        self.assertEqual(util.is_nondecreasing([-1, 1, 1, 1]), True)
        self.assertEqual(util.is_nondecreasing([1, -1, 1, 1]), False)
        self.assertEqual(util.is_nondecreasing([1, 1, -1, 1]), False)
        self.assertEqual(util.is_nondecreasing([1, 1, 1, -1]), False)

    def test_is_nonincreasing(self):
        self.assertEqual(util.is_nonincreasing([]), True)
        self.assertEqual(util.is_nonincreasing([1]), True)
        self.assertEqual(util.is_nonincreasing([2, 1]), True)
        self.assertEqual(util.is_nonincreasing([3, 2, 1]), True)
        self.assertEqual(util.is_nonincreasing([4, 3, 2, 1]), True)
        self.assertEqual(util.is_nonincreasing([3, 3, 1, 1]), True)
        self.assertEqual(util.is_nonincreasing([4, 1, 1, 1]), True)
        self.assertEqual(util.is_nonincreasing([1, 1, 1, 1]), True)
        self.assertEqual(util.is_nonincreasing([-1, 1, 1, 1]), False)
        self.assertEqual(util.is_nonincreasing([1, -1, 1, 1]), False)
        self.assertEqual(util.is_nonincreasing([1, 1, -1, 1]), False)
        self.assertEqual(util.is_nonincreasing([1, 1, 1, -1]), True)

    def test_is_positive_power_of_two(self):
        self.assertEqual(util.is_positive_power_of_two(-8), False)
        self.assertEqual(util.is_positive_power_of_two(-4), False)
        self.assertEqual(util.is_positive_power_of_two(-3), False)
        self.assertEqual(util.is_positive_power_of_two(-2), False)
        self.assertEqual(util.is_positive_power_of_two(-1), False)
        self.assertEqual(util.is_positive_power_of_two(0), False)
        self.assertEqual(util.is_positive_power_of_two(1), True)
        self.assertEqual(util.is_positive_power_of_two(2), True)
        self.assertEqual(util.is_positive_power_of_two(3), False)
        self.assertEqual(util.is_positive_power_of_two(4), True)
        self.assertEqual(util.is_positive_power_of_two(5), False)
        self.assertEqual(util.is_positive_power_of_two(6), False)
        self.assertEqual(util.is_positive_power_of_two(7), False)
        self.assertEqual(util.is_positive_power_of_two(8), True)
        self.assertEqual(util.is_positive_power_of_two(15), False)
        self.assertEqual(util.is_positive_power_of_two(16), True)
        self.assertEqual(util.is_positive_power_of_two(31), False)
        self.assertEqual(util.is_positive_power_of_two(32), True)

    def test_log2floor(self):
        self.assertEqual(util.log2floor(1), 0)
        self.assertEqual(util.log2floor(2), 1)
        self.assertEqual(util.log2floor(3), 1)
        self.assertEqual(util.log2floor(4), 2)
        self.assertEqual(util.log2floor(5), 2)
        self.assertEqual(util.log2floor(6), 2)
        self.assertEqual(util.log2floor(7), 2)
        self.assertEqual(util.log2floor(8), 3)
        self.assertEqual(util.log2floor(9), 3)
        self.assertEqual(util.log2floor(2**10), 10)
        self.assertEqual(util.log2floor(2**10 + 1), 10)
        self.assertEqual(util.log2floor(2**20), 20)
        self.assertEqual(util.log2floor(2**20 + 1), 20)
        self.assertEqual(util.log2floor(2**30), 30)
        self.assertEqual(util.log2floor(2**30 + 1), 30)
        self.assertEqual(util.log2floor(2**40), 40)
        self.assertEqual(util.log2floor(2**40 + 1), 40)

    def test_generate_gray_code(self):
        self.assertEqual(util.generate_gray_code(0), [[]])
        self.assertEqual(util.generate_gray_code(1), [[0], [1]])
        self.assertEqual(util.generate_gray_code(2), [[0, 0], [0, 1], [1, 1], [1, 0]])
        self.assertEqual(
            util.generate_gray_code(3),
            [
                [0, 0, 0],
                [0, 0, 1],
                [0, 1, 1],
                [0, 1, 0],
                [1, 1, 0],
                [1, 1, 1],
                [1, 0, 1],
                [1, 0, 0],
            ],
        )
        self.assertEqual(
            util.generate_gray_code(4),
            [
                [0, 0, 0, 0],
                [0, 0, 0, 1],
                [0, 0, 1, 1],
                [0, 0, 1, 0],
                [0, 1, 1, 0],
                [0, 1, 1, 1],
                [0, 1, 0, 1],
                [0, 1, 0, 0],
                [1, 1, 0, 0],
                [1, 1, 0, 1],
                [1, 1, 1, 1],
                [1, 1, 1, 0],
                [1, 0, 1, 0],
                [1, 0, 1, 1],
                [1, 0, 0, 1],
                [1, 0, 0, 0],
            ],
        )

    def test_characterize_function(self):
        with self.assertRaises(ValueError):
            util.characterize_function([1, 2, -1], [1, 1, 1])

        fc, slopes = util.characterize_function([1, 2, 3], [1, 1, 1])
        self.assertEqual(fc, 1)  # affine
        self.assertEqual(slopes, [0, 0])

        fc, slopes = util.characterize_function([1, 2, 3], [1, 0, 1])
        self.assertEqual(fc, 2)  # convex
        self.assertEqual(slopes, [-1, 1])

        fc, slopes = util.characterize_function([1, 2, 3], [1, 2, 1])
        self.assertEqual(fc, 3)  # concave
        self.assertEqual(slopes, [1, -1])

        fc, slopes = util.characterize_function([1, 1, 2], [1, 2, 1])
        self.assertEqual(fc, 4)  # step
        self.assertEqual(slopes, [None, -1])

        fc, slopes = util.characterize_function([1, 2, 3, 4], [1, 2, 1, 2])
        self.assertEqual(fc, 5)  # none of the above
        self.assertEqual(slopes, [1, -1, 1])

    @unittest.skipUnless(
        util.numpy_available and util.scipy_available, "Numpy or Scipy is not available"
    )
    def test_generate_delaunay(self):
        vlist = variable_list()
        vlist.append(variable(lb=0, ub=1))
        vlist.append(variable(lb=1, ub=2))
        vlist.append(variable(lb=2, ub=3))
        tri = util.generate_delaunay(vlist, num=2)
        self.assertTrue(isinstance(tri, util.scipy.spatial.Delaunay))
        self.assertEqual(len(tri.simplices), 6)
        self.assertEqual(len(tri.points), 8)

        tri = util.generate_delaunay(vlist, num=3)
        self.assertTrue(isinstance(tri, util.scipy.spatial.Delaunay))
        # we got some simplices
        self.assertTrue(len(tri.simplices) > 1)
        # all the given points are accounted for
        self.assertEqual(len(tri.points) + len(tri.coplanar), 27)

        #
        # Check cases where not all variables are bounded
        #
        vlist = variable_list()
        vlist.append(variable(lb=0))
        with self.assertRaises(ValueError):
            util.generate_delaunay(vlist)

        vlist = variable_list()
        vlist.append(variable(ub=0))
        with self.assertRaises(ValueError):
            util.generate_delaunay(vlist)


class Test_piecewise(unittest.TestCase):
    def test_pickle(self):
        for key in transforms.registered_transforms:
            v = variable(lb=1, ub=3)
            p = transforms.piecewise(
                [1, 2, 3], [1, 2, 1], input=v, validate=False, repn=key
            )
            self.assertEqual(p.parent, None)
            self.assertEqual(p.input.expr.parent, None)
            self.assertIs(p.input.expr, v)
            pup = pickle.loads(pickle.dumps(p))
            self.assertEqual(pup.parent, None)
            self.assertEqual(pup.input.expr.parent, None)
            self.assertIsNot(pup.input.expr, v)
            b = block()
            b.v = v
            b.p = p
            self.assertIs(p.parent, b)
            self.assertEqual(p.input.expr.parent, b)
            bup = pickle.loads(pickle.dumps(b))
            pup = bup.p
            self.assertIs(pup.parent, bup)
            self.assertEqual(pup.input.expr.parent, bup)
            self.assertIs(pup.input.expr, bup.v)
            self.assertIsNot(pup.input.expr, b.v)

    def test_call(self):
        g = PiecewiseLinearFunction([1], [0])
        f = TransformedPiecewiseLinearFunction(g, require_bounded_input_variable=False)
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 0)
        self.assertIs(type(f(1)), float)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(1.1)

        g = PiecewiseLinearFunction([1, 2], [0, 4])
        f = TransformedPiecewiseLinearFunction(g, require_bounded_input_variable=False)
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 0)
        self.assertIs(type(f(1)), float)
        self.assertEqual(f(1.5), 2)
        self.assertIs(type(f(1.5)), float)
        self.assertEqual(f(2), 4)
        self.assertIs(type(f(2)), float)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(2.1)

        # step function
        g = PiecewiseLinearFunction([1, 1], [0, 1])
        f = TransformedPiecewiseLinearFunction(g, require_bounded_input_variable=False)
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 0)
        self.assertIs(type(f(1)), float)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(1.1)

        g = PiecewiseLinearFunction([1, 2, 3], [1, 2, 1])
        f = TransformedPiecewiseLinearFunction(g, require_bounded_input_variable=False)
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertIs(type(f(1)), float)
        self.assertEqual(f(1.5), 1.5)
        self.assertIs(type(f(1.5)), float)
        self.assertEqual(f(2), 2)
        self.assertIs(type(f(2)), float)
        self.assertEqual(f(2.5), 1.5)
        self.assertIs(type(f(2.5)), float)
        self.assertEqual(f(3), 1)
        self.assertIs(type(f(3)), float)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(3.1)

        # step function
        g = PiecewiseLinearFunction([1, 2, 2, 3], [1, 2, 3, 4])
        f = TransformedPiecewiseLinearFunction(g, require_bounded_input_variable=False)
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertIs(type(f(1)), float)
        self.assertEqual(f(1.5), 1.5)
        self.assertIs(type(f(1.5)), float)
        self.assertEqual(f(2), 2)  # lower semicontinuous
        self.assertIs(type(f(2)), float)
        self.assertEqual(f(2.5), 3.5)
        self.assertIs(type(f(2.5)), float)
        self.assertEqual(f(3), 4)
        self.assertIs(type(f(3)), float)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(3.1)

        # another step function
        g = PiecewiseLinearFunction(
            [1, 1, 2, 3], [1, 2, 3, 4], equal_slopes_tolerance=-1
        )
        f = TransformedPiecewiseLinearFunction(
            g, require_bounded_input_variable=False, equal_slopes_tolerance=-1
        )
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(1.5), 2.5)
        self.assertEqual(f(2), 3)  # lower semicontinuous
        self.assertEqual(f(2.5), 3.5)
        self.assertEqual(f(3), 4)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(3.1)

        # another step function
        g = PiecewiseLinearFunction(
            [1, 2, 3, 3], [1, 2, 3, 4], equal_slopes_tolerance=-1
        )
        f = TransformedPiecewiseLinearFunction(
            g, require_bounded_input_variable=False, equal_slopes_tolerance=-1
        )
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(1.5), 1.5)
        self.assertEqual(f(2), 2)
        self.assertEqual(f(2.5), 2.5)
        self.assertEqual(f(3), 3)  # lower semicontinuous
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(3.1)

        # another step function using parameters
        g = PiecewiseLinearFunction(
            [pmo.parameter(1), pmo.parameter(1), pmo.parameter(2), pmo.parameter(3)],
            [pmo.parameter(1), pmo.parameter(2), pmo.parameter(3), pmo.parameter(4)],
            equal_slopes_tolerance=-1,
        )
        f = TransformedPiecewiseLinearFunction(
            g, require_bounded_input_variable=False, equal_slopes_tolerance=-1
        )
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(1.5), 2.5)
        self.assertEqual(f(2), 3)  # lower semicontinuous
        self.assertEqual(f(2.5), 3.5)
        self.assertEqual(f(3), 4)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(3.1)

        # another step function
        g = PiecewiseLinearFunction(
            [1, 1, 2, 3, 4], [1, 2, 3, 4, 5], equal_slopes_tolerance=-1
        )
        f = TransformedPiecewiseLinearFunction(
            g, require_bounded_input_variable=False, equal_slopes_tolerance=-1
        )
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)  # lower semicontinuous
        self.assertEqual(f(1.5), 2.5)
        self.assertEqual(f(2), 3)
        self.assertEqual(f(2.5), 3.5)
        self.assertEqual(f(3), 4)
        self.assertEqual(f(3.5), 4.5)
        self.assertEqual(f(4), 5)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(4.1)

        # another step function
        g = PiecewiseLinearFunction(
            [1, 2, 2, 3, 4], [1, 2, 3, 4, 5], equal_slopes_tolerance=-1
        )
        f = TransformedPiecewiseLinearFunction(
            g, require_bounded_input_variable=False, equal_slopes_tolerance=-1
        )
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(1.5), 1.5)
        self.assertEqual(f(2), 2)  # lower semicontinuous
        self.assertEqual(f(2.5), 3.5)
        self.assertEqual(f(3), 4)
        self.assertEqual(f(3.5), 4.5)
        self.assertEqual(f(4), 5)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(4.1)

        # another step function
        g = PiecewiseLinearFunction(
            [1, 2, 3, 3, 4], [1, 2, 3, 4, 5], equal_slopes_tolerance=-1
        )
        f = TransformedPiecewiseLinearFunction(
            g, require_bounded_input_variable=False, equal_slopes_tolerance=-1
        )
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(1.5), 1.5)
        self.assertEqual(f(2), 2)
        self.assertEqual(f(2.5), 2.5)
        self.assertEqual(f(3), 3)  # lower semicontinuous
        self.assertEqual(f(3.5), 4.5)
        self.assertEqual(f(4), 5)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(4.1)

        # another step function
        g = PiecewiseLinearFunction(
            [1, 2, 3, 4, 4], [1, 2, 3, 4, 5], equal_slopes_tolerance=-1
        )
        f = TransformedPiecewiseLinearFunction(
            g, require_bounded_input_variable=False, equal_slopes_tolerance=-1
        )
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(1.5), 1.5)
        self.assertEqual(f(2), 2)
        self.assertEqual(f(2.5), 2.5)
        self.assertEqual(f(3), 3)
        self.assertEqual(f(3.5), 3.5)
        self.assertEqual(f(4), 4)  # lower semicontinuous
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(4.1)

        # another step function
        g = PiecewiseLinearFunction(
            [1, 2, 3, 4, 5], [1, 2, 3, 4, 5], equal_slopes_tolerance=-1
        )
        f = TransformedPiecewiseLinearFunction(
            g, require_bounded_input_variable=False, equal_slopes_tolerance=-1
        )
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(1.5), 1.5)
        self.assertEqual(f(2), 2)
        self.assertEqual(f(2.5), 2.5)
        self.assertEqual(f(3), 3)
        self.assertEqual(f(3.5), 3.5)
        self.assertEqual(f(4), 4)
        self.assertEqual(f(4.5), 4.5)
        self.assertEqual(f(5), 5)
        with self.assertRaises(ValueError):
            f(0.9)
        with self.assertRaises(ValueError):
            f(5.1)

    def test_type(self):
        for key in transforms.registered_transforms:
            p = transforms.piecewise([1, 2, 3], [1, 2, 1], repn=key, validate=False)
            self.assertTrue(len(list(p.children())) <= 4)
            self.assertTrue(isinstance(p, TransformedPiecewiseLinearFunction))
            self.assertTrue(isinstance(p, transforms.registered_transforms[key]))
            self.assertTrue(isinstance(p, ICategorizedObject))
            self.assertTrue(isinstance(p, ICategorizedObjectContainer))
            self.assertTrue(isinstance(p, IHeterogeneousContainer))
            self.assertTrue(isinstance(p, IBlock))
            self.assertTrue(isinstance(p, block))

    def test_bad_repn(self):
        repn = list(transforms.registered_transforms.keys())[0]
        self.assertTrue(repn in transforms.registered_transforms)
        transforms.piecewise([1, 2, 3], [1, 2, 1], validate=False, repn=repn)

        repn = '_bad_repn_'
        self.assertFalse(repn in transforms.registered_transforms)
        with self.assertRaises(ValueError):
            transforms.piecewise([1, 2, 3], [1, 2, 1], validate=False, repn=repn)
        with self.assertRaises(ValueError):
            transforms.piecewise(
                [1, 2, 3],
                [1, 2, 1],
                input=variable(lb=1, ub=3),
                validate=True,
                simplify=False,
                repn=repn,
            )
        with self.assertRaises(ValueError):
            transforms.piecewise(
                [1, 2, 3],
                [1, 2, 1],
                input=variable(lb=1, ub=3),
                validate=True,
                simplify=True,
                repn=repn,
            )

    def test_init(self):
        for key in transforms.registered_transforms:
            for bound in ['lb', 'ub', 'eq', 'bad']:
                for args in [
                    ([1, 2, 3], [1, 2, 1]),
                    ([1, 2, 3, 4, 5], [1, 2, 1, 2, 1]),
                    ([1, 2, 3, 4, 5, 6, 7, 8, 9], [1, 2, 1, 2, 1, 2, 1, 2, 1]),
                ]:
                    kwds = {'repn': key, 'bound': bound, 'validate': False}
                    if bound == 'bad':
                        with self.assertRaises(ValueError):
                            transforms.piecewise(*args, **kwds)
                        kwds['simplify'] = True
                        with self.assertRaises(ValueError):
                            transforms.piecewise(*args, **kwds)
                        kwds['simplify'] = False
                        with self.assertRaises(ValueError):
                            transforms.piecewise(*args, **kwds)
                    else:
                        p = transforms.piecewise(*args, **kwds)
                        self.assertTrue(
                            isinstance(p, transforms.registered_transforms[key])
                        )
                        self.assertTrue(
                            isinstance(p, TransformedPiecewiseLinearFunction)
                        )
                        self.assertEqual(p.active, True)
                        self.assertIs(p.parent, None)
                        kwds['simplify'] = True
                        p = transforms.piecewise(*args, **kwds)
                        self.assertTrue(
                            isinstance(p, transforms.registered_transforms[key])
                        )
                        self.assertTrue(
                            isinstance(p, TransformedPiecewiseLinearFunction)
                        )
                        self.assertEqual(p.active, True)
                        self.assertIs(p.parent, None)
                        kwds['simplify'] = False
                        p = transforms.piecewise(*args, **kwds)
                        self.assertTrue(
                            isinstance(p, transforms.registered_transforms[key])
                        )
                        self.assertTrue(
                            isinstance(p, TransformedPiecewiseLinearFunction)
                        )
                        self.assertEqual(p.active, True)
                        self.assertIs(p.parent, None)

    def test_bad_init(self):
        # lists not the same length
        with self.assertRaises(ValueError):
            PiecewiseLinearFunction([1, 2, 3], [1, 2, 1, 1], validate=False)
        # lists not the same length
        with self.assertRaises(ValueError):
            PiecewiseLinearFunction([1, 2, 3, 4], [1, 2, 1], validate=False)

        # breakpoints list not nondecreasing
        with self.assertRaises(util.PiecewiseValidationError):
            PiecewiseLinearFunction([1, 3, 2], [1, 2, 1])

        PiecewiseLinearFunction([1, 3, 2], [1, 2, 1], validate=False)

        PiecewiseLinearFunction(
            [1, 2, 3], [1, 1, 1 + 2e-6], equal_slopes_tolerance=1e-6
        )

        # consecutive slopes are "equal"
        with self.assertRaises(util.PiecewiseValidationError):
            PiecewiseLinearFunction(
                [1, 2, 3], [1, 1, 1 + 2e-6], equal_slopes_tolerance=3e-6
            )

        PiecewiseLinearFunction([1, 2, 3], [1, 1, 1 + 2e-6], validate=False)

        f = PiecewiseLinearFunction([1, 2, 3], [1, 2, 1])
        TransformedPiecewiseLinearFunction(
            f, input=variable(lb=1, ub=3), require_bounded_input_variable=True
        )

        TransformedPiecewiseLinearFunction(
            f, input=variable(lb=1, ub=3), require_bounded_input_variable=False
        )

        # variable is not bounded
        with self.assertRaises(util.PiecewiseValidationError):
            TransformedPiecewiseLinearFunction(
                f, input=variable(lb=1), require_bounded_input_variable=True
            )
        TransformedPiecewiseLinearFunction(
            f, input=variable(lb=1), require_bounded_input_variable=False
        )
        with self.assertRaises(util.PiecewiseValidationError):
            TransformedPiecewiseLinearFunction(
                f, input=variable(ub=3), require_bounded_input_variable=True
            )
        TransformedPiecewiseLinearFunction(
            f, input=variable(ub=3), require_bounded_input_variable=False
        )
        with self.assertRaises(util.PiecewiseValidationError):
            TransformedPiecewiseLinearFunction(f, require_bounded_input_variable=True)
        TransformedPiecewiseLinearFunction(f, require_bounded_input_variable=False)

        # variable domain is not fully covered
        with self.assertRaises(util.PiecewiseValidationError):
            TransformedPiecewiseLinearFunction(
                f,
                input=variable(lb=0),
                require_bounded_input_variable=False,
                require_variable_domain_coverage=True,
            )
        TransformedPiecewiseLinearFunction(
            f,
            input=variable(lb=0),
            require_bounded_input_variable=False,
            require_variable_domain_coverage=False,
        )
        with self.assertRaises(util.PiecewiseValidationError):
            TransformedPiecewiseLinearFunction(
                f,
                input=variable(ub=4),
                require_bounded_input_variable=False,
                require_variable_domain_coverage=True,
            )
        TransformedPiecewiseLinearFunction(
            f,
            input=variable(ub=4),
            require_bounded_input_variable=False,
            require_variable_domain_coverage=False,
        )

    def test_bad_init_log_types(self):
        # lists are not of length: (2^n) + 1
        with self.assertRaises(ValueError):
            transforms.piecewise(
                [1, 2, 3, 4], [1, 2, 3, 4], repn='dlog', validate=False
            )
        with self.assertRaises(ValueError):
            transforms.piecewise([1, 2, 3, 4], [1, 2, 3, 4], repn='log', validate=False)

    def test_step(self):
        breakpoints = [1, 2, 2]
        values = [1, 0, 1]
        v = variable()
        v.bounds = min(breakpoints), max(breakpoints)
        for key in transforms.registered_transforms:
            if key in ('mc', 'convex'):
                with self.assertRaises(util.PiecewiseValidationError):
                    transforms.piecewise(breakpoints, values, input=v, repn=key)
            else:
                p = transforms.piecewise(breakpoints, values, input=v, repn=key)
                self.assertEqual(p.validate(), 4)

    def test_simplify(self):
        v = variable(lb=1, ub=3)
        convex_breakpoints = [1, 2, 3]
        convex_values = [1, 0, 1]
        for key in transforms.registered_transforms:
            for bound in ('lb', 'ub', 'eq'):
                if (key == 'convex') and (bound != 'lb'):
                    with self.assertRaises(util.PiecewiseValidationError):
                        transforms.piecewise(
                            convex_breakpoints,
                            convex_values,
                            input=v,
                            repn=key,
                            bound=bound,
                            simplify=False,
                        )
                    with self.assertRaises(util.PiecewiseValidationError):
                        transforms.piecewise(
                            convex_breakpoints,
                            convex_values,
                            input=v,
                            repn=key,
                            bound=bound,
                            simplify=True,
                        )
                else:
                    p = transforms.piecewise(
                        convex_breakpoints,
                        convex_values,
                        input=v,
                        repn=key,
                        bound=bound,
                        simplify=False,
                    )
                    self.assertTrue(
                        isinstance(p, transforms.registered_transforms[key])
                    )
                    self.assertEqual(p.validate(), util.characterize_function.convex)
                    p = transforms.piecewise(
                        convex_breakpoints,
                        convex_values,
                        input=v,
                        repn=key,
                        bound=bound,
                        simplify=True,
                    )
                    if bound == 'lb':
                        self.assertTrue(
                            isinstance(p, transforms.registered_transforms['convex'])
                        )
                    else:
                        self.assertTrue(
                            isinstance(p, transforms.registered_transforms[key])
                        )

        concave_breakpoints = [1, 2, 3]
        concave_values = [-1, 0, -1]
        for key in transforms.registered_transforms:
            for bound in ('lb', 'ub', 'eq'):
                if (key == 'convex') and (bound != 'ub'):
                    with self.assertRaises(util.PiecewiseValidationError):
                        transforms.piecewise(
                            concave_breakpoints,
                            concave_values,
                            input=v,
                            repn=key,
                            bound=bound,
                            simplify=False,
                        )
                    with self.assertRaises(util.PiecewiseValidationError):
                        transforms.piecewise(
                            concave_breakpoints,
                            concave_values,
                            input=v,
                            repn=key,
                            bound=bound,
                            simplify=True,
                        )
                else:
                    p = transforms.piecewise(
                        concave_breakpoints,
                        concave_values,
                        input=v,
                        repn=key,
                        bound=bound,
                        simplify=False,
                    )
                    self.assertTrue(
                        isinstance(p, transforms.registered_transforms[key])
                    )
                    self.assertEqual(p.validate(), util.characterize_function.concave)
                    p = transforms.piecewise(
                        concave_breakpoints,
                        concave_values,
                        input=v,
                        repn=key,
                        bound=bound,
                        simplify=True,
                    )
                    if bound == 'ub':
                        self.assertTrue(
                            isinstance(p, transforms.registered_transforms['convex'])
                        )
                    else:
                        self.assertTrue(
                            isinstance(p, transforms.registered_transforms[key])
                        )

        affine_breakpoints = [1, 3]
        affine_values = [1, 3]
        for key in transforms.registered_transforms:
            for bound in ('lb', 'ub', 'eq'):
                p = transforms.piecewise(
                    affine_breakpoints,
                    affine_values,
                    input=v,
                    repn=key,
                    bound=bound,
                    simplify=False,
                )
                self.assertTrue(isinstance(p, transforms.registered_transforms[key]))
                self.assertEqual(p.validate(), util.characterize_function.affine)
                p = transforms.piecewise(
                    affine_breakpoints,
                    affine_values,
                    input=v,
                    repn=key,
                    bound=bound,
                    simplify=True,
                )
                self.assertTrue(
                    isinstance(p, transforms.registered_transforms['convex'])
                )


class Test_piecewise_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = block_dict
    _ctype_factory = lambda self: transforms.piecewise(
        [1, 2, 3], [1, 2, 1], validate=False
    )


class Test_piecewise_list(_TestActiveListContainerBase, unittest.TestCase):
    _container_type = block_list
    _ctype_factory = lambda self: transforms.piecewise(
        [1, 2, 3], [1, 2, 1], validate=False
    )


@unittest.skipUnless(
    util.numpy_available and util.scipy_available, "Numpy or Scipy is not available"
)
class Test_piecewise_nd(unittest.TestCase):
    def test_pickle(self):
        for key in transforms_nd.registered_transforms:
            p = transforms_nd.piecewise_nd(_test_tri, _test_values, repn=key)
            self.assertEqual(p.parent, None)
            pup = pickle.loads(pickle.dumps(p))
            self.assertEqual(pup.parent, None)
            b = block()
            b.p = p
            self.assertIs(p.parent, b)
            bup = pickle.loads(pickle.dumps(b))
            pup = bup.p
            self.assertIs(pup.parent, bup)

    def test_call(self):
        #
        # 2d points
        #
        vlist = variable_list([variable(lb=0, ub=1), variable(lb=0, ub=1)])
        tri = util.generate_delaunay(vlist, num=3)
        x, y = tri.points.T
        values = x * y
        g = PiecewiseLinearFunctionND(tri, values)
        f = TransformedPiecewiseLinearFunctionND(g)
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertTrue(util.numpy.isclose(f(tri.points), values).all())
        self.assertAlmostEqual(f([0, 0]), 0.0)
        self.assertAlmostEqual(f(util.numpy.array([0, 0])), 0.0)
        self.assertAlmostEqual(f([1, 1]), 1.0)
        self.assertAlmostEqual(f(util.numpy.array([1, 1])), 1.0)

        #
        # 3d points
        #
        vlist = variable_list(
            [variable(lb=0, ub=1), variable(lb=0, ub=1), variable(lb=0, ub=1)]
        )
        tri = util.generate_delaunay(vlist, num=10)
        x, y, z = tri.points.T
        values = x * y * z
        g = PiecewiseLinearFunctionND(tri, values)
        f = TransformedPiecewiseLinearFunctionND(g)
        self.assertTrue(f.parent is None)
        self.assertEqual(f.ctype, IBlock)
        self.assertTrue(util.numpy.isclose(f(tri.points), values).all())
        self.assertAlmostEqual(f([0, 0, 0]), 0.0)
        self.assertAlmostEqual(f(util.numpy.array([0, 0, 0])), 0.0)
        self.assertAlmostEqual(f([1, 1, 1]), 1.0)
        self.assertAlmostEqual(f(util.numpy.array([1, 1, 1])), 1.0)

    def test_type(self):
        for key in transforms_nd.registered_transforms:
            p = transforms_nd.piecewise_nd(_test_tri, _test_values, repn=key)
            # small block storage
            self.assertTrue(len(list(p.children())) <= 4)
            self.assertTrue(isinstance(p, TransformedPiecewiseLinearFunctionND))
            self.assertTrue(isinstance(p, transforms_nd.registered_transforms[key]))
            self.assertTrue(isinstance(p, ICategorizedObject))
            self.assertTrue(isinstance(p, ICategorizedObjectContainer))
            self.assertTrue(isinstance(p, IHeterogeneousContainer))
            self.assertTrue(isinstance(p, IBlock))
            self.assertTrue(isinstance(p, block))

    def test_bad_repn(self):
        repn = list(transforms_nd.registered_transforms.keys())[0]
        self.assertTrue(repn in transforms_nd.registered_transforms)
        transforms_nd.piecewise_nd(_test_tri, _test_values, repn=repn)

        repn = '_bad_repn_'
        self.assertFalse(repn in transforms_nd.registered_transforms)
        with self.assertRaises(ValueError):
            transforms_nd.piecewise_nd(_test_tri, _test_values, repn=repn)

    def test_init(self):
        for key in transforms_nd.registered_transforms:
            for bound in ['lb', 'ub', 'eq', 'bad']:
                args = (_test_tri, _test_values)
                kwds = {'repn': key, 'bound': bound}
                if bound == 'bad':
                    with self.assertRaises(ValueError):
                        transforms_nd.piecewise_nd(*args, **kwds)
                else:
                    p = transforms_nd.piecewise_nd(*args, **kwds)
                    self.assertTrue(
                        isinstance(p, transforms_nd.registered_transforms[key])
                    )
                    self.assertTrue(isinstance(p, TransformedPiecewiseLinearFunctionND))
                    self.assertEqual(p.active, True)
                    self.assertIs(p.parent, None)


@unittest.skipUnless(
    util.numpy_available and util.scipy_available, "Numpy or Scipy is not available"
)
class Test_piecewise_nd_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = block_dict
    _ctype_factory = lambda self: transforms_nd.piecewise_nd(_test_tri, _test_values)


@unittest.skipUnless(
    util.numpy_available and util.scipy_available, "Numpy or Scipy is not available"
)
class Test_piecewise_nd_list(_TestActiveListContainerBase, unittest.TestCase):
    _container_type = block_list
    _ctype_factory = lambda self: transforms_nd.piecewise_nd(_test_tri, _test_values)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""A library of possible callbacks to use for initializing the NLP subproblems.
However, it seems likely that problem-specific routines might be better, in
which case you can write your own, and specify it in the
'subproblem_initialization_method' argument."""


def _restore_vars_from_nlp_block_saved_values(nlp_util_block):
    for var, old_value in nlp_util_block.initial_var_values.items():
        if not var.fixed and var.is_continuous():
            if old_value is not None:
                # Adjust value if it falls outside the bounds
                if var.has_lb() and old_value < var.lb:
                    old_value = var.lb
                if var.has_ub() and old_value > var.ub:
                    old_value = var.ub
                # Set the value
                var.set_value(old_value)


# This is the original GDPopt behavior:
def restore_vars_to_original_values(solver, nlp_util_block, mip_util_block):
    """Perform initialization of the subproblem.

    This just restores the continuous variables to the original
    model values, which were saved on the subproblem's utility block when it
    was created.
    """
    # restore original continuous variable values
    _restore_vars_from_nlp_block_saved_values(nlp_util_block)


# This is the default for the enumerate algorithm
def restore_vars_to_original_values_enumerate(
    true_disjuncts, boolean_var_values, discrete_var_values, nlp_util_block
):
    """Perform initialization of the subproblem.

    This just restores the continuous variables to the original
    model values, which were saved on the subproblem's utility block when it
    was created.
    """
    # restore original continuous variable values
    _restore_vars_from_nlp_block_saved_values(nlp_util_block)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
from pyomo.environ import (
    SolverFactory,
    ConcreteModel,
    Var,
    Constraint,
    Objective,
    Integers,
    Boolean,
    Suffix,
    maximize,
)
from pyomo.common.tee import capture_output
from pyomo.common.tempfiles import TempfileManager
import pyomo.common.unittest as unittest

opt_cbc = SolverFactory('cbc')
cbc_available = opt_cbc.available(exception_flag=False)


@unittest.pytest.mark.solver("cbc")
class CBCTests(unittest.TestCase):
    @unittest.skipIf(not cbc_available, "The CBC solver is not available")
    def test_warm_start(self):
        m = ConcreteModel()
        m.x = Var()
        m.z = Var(domain=Integers)
        m.w = Var(domain=Boolean)
        m.c = Constraint(expr=m.x + m.z + m.w >= 0)
        m.o = Objective(expr=m.x + m.z + m.w)

        TempfileManager.push()
        tempdir = os.path.dirname(TempfileManager.create_tempfile())
        TempfileManager.pop()

        sameDrive = os.path.splitdrive(tempdir)[0] == os.path.splitdrive(os.getcwd())[0]

        # At the moment, CBC does not cleanly handle windows-style drive
        # names in the MIPSTART file name (though at least 2.10.5).
        #
        # See https://github.com/coin-or/Cbc/issues/32
        # The problematic source is https://github.com/coin-or/Cbc/blob/3dcedb27664ae458990e9d4d50bc11c2c55917a0/src/CbcSolver.cpp#L9445-L9459
        #
        # We will try two different situations: running from the current
        # directory (which may or may not be on the same drive), and
        # then from the tempdir (which will be on the same drive).

        # Set some initial values for warm start.
        m.x.set_value(10)
        m.z.set_value(5)
        m.w.set_value(1)

        with SolverFactory("cbc") as opt, capture_output() as output:
            opt.solve(
                m, tee=True, warmstart=True, options={'sloglevel': 2, 'loglevel': 2}
            )

        log = output.getvalue()
        # Check if CBC loaded the warmstart file.
        self.assertIn('opening mipstart file', log)

        if sameDrive:
            # Only integer and binary variables are considered by CBC.
            self.assertIn('MIPStart values read for 2 variables.', log)
            # m.x is ignored because it is continuous, so cost should be 5+1
            self.assertIn('MIPStart provided solution with cost 6', log)
        else:
            self.assertNotIn('MIPStart values read', log)

        # Set some initial values for warm start.
        m.x.set_value(10)
        m.z.set_value(5)
        m.w.set_value(1)

        try:
            _origDir = os.getcwd()
            os.chdir(tempdir)
            with SolverFactory("cbc") as opt, capture_output() as output:
                opt.solve(
                    m, tee=True, warmstart=True, options={'sloglevel': 2, 'loglevel': 2}
                )
        finally:
            os.chdir(_origDir)

        log = output.getvalue()
        # Check if CBC loaded the warmstart file.
        self.assertIn('opening mipstart file', log)
        # Only integer and binary variables are considered by CBC.
        self.assertIn('MIPStart values read for 2 variables.', log)
        # m.x is ignored because it is continuous, so cost should be 5+1
        self.assertIn('MIPStart provided solution with cost 6', log)

    @unittest.skipIf(not cbc_available, "The CBC solver is not available")
    def test_duals_signs(self):
        m = ConcreteModel()
        m.x = Var()
        m.obj = Objective(expr=m.x)
        m.c = Constraint(expr=(-1, m.x, 1))
        m.dual = Suffix(direction=Suffix.IMPORT)
        opt = SolverFactory('cbc')
        res = opt.solve(m)
        self.assertAlmostEqual(res.problem.lower_bound, -1)
        self.assertAlmostEqual(res.problem.upper_bound, -1)
        self.assertAlmostEqual(m.dual[m.c], 1)
        m.obj.sense = maximize
        res = opt.solve(m)
        self.assertAlmostEqual(res.problem.lower_bound, 1)
        self.assertAlmostEqual(res.problem.upper_bound, 1)
        self.assertAlmostEqual(m.dual[m.c], 1)

    @unittest.skipIf(not cbc_available, "The CBC solver is not available")
    def test_rc_signs(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-1, 1))
        m.obj = Objective(expr=m.x)
        m.rc = Suffix(direction=Suffix.IMPORT)
        opt = SolverFactory('cbc')
        res = opt.solve(m)
        self.assertAlmostEqual(res.problem.lower_bound, -1)
        self.assertAlmostEqual(res.problem.upper_bound, -1)
        self.assertAlmostEqual(m.rc[m.x], 1)
        m.obj.sense = maximize
        res = opt.solve(m)
        self.assertAlmostEqual(res.problem.lower_bound, 1)
        self.assertAlmostEqual(res.problem.upper_bound, 1)
        self.assertAlmostEqual(m.rc[m.x], 1)


if __name__ == "__main__":
    unittest.main()
